
/*
Copyright 2007, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2007
*/

#include <ScatterWeb.System.h>
#include <ScatterWeb.CC1020.h>
#include <string.h>
#include <stdio.h>
#include "dsr.h"
#include "transport_layer.h"
#include "ScatterWeb.Process.h"



bool Process_radioHandler(struct netpacket_handler_args* args)
{
	/*
		Notes: * The received packet data becomes invalid when this handler is completed.
	*/

	// Pass packet through command interface
	if ( Cmd_radioHandler( args ) ) {
		// packet handled by messaging module
		return true;
	}

	//check if received packet belongs to local network range
	if( !IsNtwAddr(args->netheader->to)) return false;
	else if (args->netheader->to == LOCAL_BROADCAST){

		// process local broadcast request packets here
		switch ( args->netheader->type ) {
			case DSR_REQUEST_PACKET:
				// handle dsr request packet
				if (DEBUGDSR) printf("I have received a DSR Request Packet!");
				Handle_Req_Pckt((req_packet_t*) args->payload);
				break;

			default:
				return false;	
		}

	}

	if ( args->cpaResult == ADRCOMP_OTHER ) {
		// intercept packets for other nodes here

	} else if ( args->cpaResult == ADRCOMP_BROADCAST ) {

	} else {
		// process packets for this node here
		switch ( args->netheader->type ) {
			case PING_REQUEST_PACKET:
				// send back to receiver's address
				Net_sendPong(args->netheader->from);
				break;

			case PING_REPLY_PACKET:
				LOG_LOW( "PONG from %u", args->netheader->from );
				break;

			case DSR_RESPONSE_PACKET:
				if(DEBUGDSR) printf("I have received a DSR response packet!");
				Handle_Rsp_Pckt((rsp_packet_t*) args->payload);
				break;

			case DSR_ERROR_PACKET:
				if(DEBUGDSR) printf("I have received an error packet!");
				break;

			case DSR_DATA_PACKET:
				if(DEBUGDSR) printf("I have received a data packet!");
				Handle_Data_Pckt((data_packet_t*) args->payload);
				break;

			case TCP_DATA_PACKET:
				if(DEBUGDSR) printf("I have received a transport data packet!");
				Handle_Tcp_Data_Pckt((tcp_packet_t*) args->payload, args->netheader->from);

			default:
				return false;		// packet not handled (unknown)
		}
		return true;				// packet handled
	}
	return false;					// packet not handled (other destination)
}

void Process_sensorHandler(void* sdata)
{

}

void Process_init()
{
	System_registerCallback( C_RADIO, ( fp_vp )Process_radioHandler );
	System_registerCallback( C_SENSOR, ( fp_vp )Process_sensorHandler );

	//DSR Data cache & Request Table initialization 
	InitDataCache();
	InitReqTable();

	//TCP initialization
	InitSendBuffer();

	//initialize test metrics
	SendDataCtr = 0;
	SentDataPcktCtr = 0;
	RcvDataPcktCtr = 0;
	SentTCPDataCtr = 0;
	RcvTCPDataCtr = 0;
	RetxTCPDataCtr = 0;
	DropTCPDataCtr = 0;
}

/**
 * @brief	Ping request
 *
 * Send a ping request to another node
 * Example: ping 24
 */
COMMAND( ping, 0, cmdargs )
{
	uint16_t node;

	node = String_parseUint16( cmdargs->args, NULL );
	if ( node > 0 ) {
		Net_sendPing((netaddr_t)node);
	}
}

/**
 * @brief Echo Arguments
 *
 * Testcase 1: long command name
 * Testcase 2: argument splitting
 */


COMMAND( echoargs, CMDFLAG_SERIAL, cmdargs )
{
	uint16_t i;
	uint16_t count = Cmd_SplitArgs( cmdargs->args );
	char* arg = (char*) cmdargs->args;

	for ( i = 1; i <= count; i++ ) {
		printf("%u: %*s\n", i, strlen(arg), arg);
		arg = ( char* )Cmd_GetNextArg( arg );
	}
	snprintf(cmdargs->response_args, cmdargs->response_bufsize, "%u args found", count );
}

COMMAND(HelloMSB, CMDFLAG_SERIAL, cmdargs) {
	uint16_t n = String_parseUint16(cmdargs->args, NULL);
	while(n--) {
		printf("Hello User - called %u times\r\n", n);
	}
}



COMMAND(SendDsr, CMDFLAG_SERIAL, cmdargs) {
	uint16_t num_arg = 1;
	uint16_t count = Cmd_SplitArgs( cmdargs->args );

	if (count > num_arg){
		printf("ERROR: too many arguments\n");
	}
	else{
		uint16_t dest = String_parseUint16(cmdargs->args, NULL);
		printf("Destination Node: %d\n",dest);

		payload_data data = "PAYLOAD";

		data_entry_t* entry;

		//insert data entry in data cache
		entry = Insert_Data_Entry(dest, data);
		
		if (entry != 0){

			printf("Entry created in data cache\n");
			Show_Data_Cache();
			Start_Route_Discovery((netaddr_t)dest);
			Timers_add(1024*3, Restart_Route_Disc, entry);

		}
		else{//if no space available in data cache, discard entry
			printf("ERROR: no space available in data cache!\n");
		}
	}
}

void traffic_generator(void *datap) {
	
	uint16_t dstnode = (uint16_t) datap;
	char SentChar [PayloadSize - 7];

	SentDataPcktCtr++;
	printf("Sending data packet %d\n",SentDataPcktCtr);
	payload_data data = "PAYLOAD";
	itoa(SentDataPcktCtr,SentChar,10);

	//add packet number to payload data
	strcat(data, SentChar);
	data_entry_t* entry;

	//insert data entry in data cache
	entry = Insert_Data_Entry(dstnode, data);
		
	if (entry != 0){

		if (DEBUGDSR){ 
			printf("Entry created in data cache\n");
			Show_Data_Cache();
		}

		//check if dest node is already in Data Buffer
		if(isAddrInReqTable(dstnode)){
			printf("Dst node already in Req Table. No Route Discovery initialized\n");
			Show_Req_Table();
		}
		else{
			printf("Dst node isn't in Req Table. Start Route Discovery\n");
			Insert_ReqTable_Entry(dstnode);
			Show_Req_Table();
			Start_Route_Discovery((netaddr_t)dstnode);
			Timers_add(1024*RTODSR, Restart_Route_Disc, entry);
		}
	}
	else{//if no space available in data cache, discard entry
		printf("ERROR: no space available in data cache!\n");
	}

	counter--;
	if(counter > 0){
		Timers_add(interval*256, traffic_generator, dstnode);
	}
	
}


COMMAND(TestDsr, CMDFLAG_SERIAL, cmdargs) {
	uint16_t num_arg = 3;
	uint16_t count = Cmd_SplitArgs( cmdargs->args );
	uint16_t dstnode;
	

	if (count != num_arg){
		printf("ERROR: argument number mismatch\n");
	}
	else{
		char* arg = (char*) cmdargs->args;
		dstnode = String_parseUint16(cmdargs->args, NULL);
		printf("Destination Node: %d\n",dstnode);
		arg = ( char* )Cmd_GetNextArg( arg );
		counter = String_parseUint16(arg, NULL);
		printf("Num packets: %d\n",counter);
		arg = ( char* )Cmd_GetNextArg( arg );
		interval = String_parseUint16(arg, NULL);
		printf("Interpacket time: %d\n",interval);
	
		//initialize Sent and Received Data Packet counters
		SentDataPcktCtr = 0;
		RcvDataPcktCtr = 0;

		//call traffic generator function
		traffic_generator(dstnode);
	}
}

void TCPtraffic_generator_single(void *datap) {
	uint16_t dstnode = (uint16_t) datap;

	char SentChar [TCPPayloadSize - 7];

	SentDataPcktCtr++;
	printf("Sending data segment %d\n",SentDataPcktCtr);

	tcp_payload data = "PAYLOAD";

	itoa(SentDataPcktCtr,SentChar,10);

	//add packet number to payload data
	strcat(data, SentChar);
	sendbuffer_entry_t* entry;

	//check if there is a TCB created for the given destination node
	if (!isConnectionInTCBList(dstnode)){
		printf("ERROR: no TCB connection to node %d\n", dstnode);
	}
	else{
		
		//insert data entry in Send Buffer
		entry = Insert_Entry_SendBuffer(dstnode, data);
		
		if (entry != 0){

			printf("Entry created in Send Buffer\n");
			ShowSendBuffer();

			//check if connection has been established and transmission window size is greater than 1
			if((getTCBState(dstnode) == ESTAB) && (getTCBTxWndw(dstnode)>0)){
				//set segment SEQ number
				entry->SEQ_number = getTCBSEQ(dstnode);
				
				//send packet
				if (SendTCP(entry)) printf("Segment %d sent to node %d\n",SentDataPcktCtr,dstnode);

				//increment SEQ number for the connection
				updateTCBSEQ(dstnode);

				Timers_add(RTOTCP*1024, Restart_SendTCP, entry);

				//decrement Tx window size for the connection
				setTCBTxWndw(dstnode,getTCBTxWndw(dstnode)-1);
				ShowTCBList();
			}
			else{
				printf("Connection request to %d in progress...\n",dstnode);
				Timers_add(RetryTCPSend_timeout*1024, RetryTCPSend, entry);
			}
		}
		else{//if no space available in send buffer, discard entry
			printf("ERROR: no space available in send buffer!\n");
		}
	}

}


COMMAND(SendTcp, CMDFLAG_SERIAL, cmdargs) {
	uint16_t num_arg = 1;
	uint16_t count = Cmd_SplitArgs( cmdargs->args );

	if (count > num_arg){
		printf("ERROR: too many arguments\n");
	}
	else{
		uint16_t dest = String_parseUint16(cmdargs->args, NULL);
		printf("Destination Node: %d\n",dest);

		//open connection to destination node
		OpenTCP(dest); 
		
		printf("Connection establishment sent to node %d\n",dest);

		//call traffic generator function after 3 secs
		Timers_add(8*1024, TCPtraffic_generator_single, dest);
	}
}


void TCPtraffic_generator(void *datap) {
	uint16_t dstnode = (uint16_t) datap;

	char SentChar [TCPPayloadSize - 7];

	SendDataCtr++;
	printf("Sending data segment %d\n",SendDataCtr);

	tcp_payload data = "PAYLOAD";

	itoa(SendDataCtr,SentChar,10);

	//add packet number to payload data
	strcat(data, SentChar);
	sendbuffer_entry_t* entry;

	//check if there is a TCB created for the given destination node
	if (!isConnectionInTCBList(dstnode)){
		printf("ERROR: no TCB connection to node %d\n", dstnode);
	}
	else{
		
		//insert data entry in Send Buffer
		entry = Insert_Entry_SendBuffer(dstnode, data);
		
		if (entry != 0){

			printf("Entry created in Send Buffer\n");
			ShowSendBuffer();

			//check if connection has been established and 
			if(getTCBState(dstnode) == ESTAB){

				//check if transmission window size >= 1
				if((getTCBTxWndw(dstnode)>0)&&(entry->sgmnmbr == getTCBSegmNumbr(dstnode))){

					//set segment SEQ number
					entry->SEQ_number = getTCBSEQ(dstnode);

					//send packet
					if (SendTCP(entry)) printf("Segment %d: %s sent to node %d\n",SendDataCtr,entry->data,dstnode);

					incTCBSegmNumbr(dstnode);

					//increment number of sent TCP segments
					SentTCPDataCtr++;

					//increment SEQ number for the connection
					updateTCBSEQ(dstnode);

					Timers_add(RTOTCP*1024, Restart_SendTCP, entry);

					//decrement Tx window size for the connection
					setTCBTxWndw(dstnode,getTCBTxWndw(dstnode)-1);
					ShowTCBList();
				}
				else{
					printf("Transmission window size currently empty...\n",dstnode);
					ShowTCBList();
					ShowSendBuffer();
					if (!LookupWaitingSegment(dstnode)) printf("No data waiting to be sent!!\n");
					printf("NextSegment: %d\n", getTCBSegmNumbr(dstnode));
				}
			}
			else{
				printf("Connection request to %d in progress...\n",dstnode);
			}
		}
		else{//if no space available in send buffer, discard entry
			printf("ERROR: Send Buffer overflow. Segment dropped!\n");
			DropTCPDataCtr ++;
		}
	}

	counter--;
	if(counter > 0){
		Timers_add(interval*256, TCPtraffic_generator, dstnode);
	}
	else{
		//call close connection function after 10 secs
		Timers_add(10*1024, CloseTCP, dstnode);
	}
}



COMMAND(TestTcp, CMDFLAG_SERIAL, cmdargs) {
	uint16_t num_arg = 3;
	uint16_t count = Cmd_SplitArgs( cmdargs->args );
	uint16_t dstnode;

	if (count != num_arg){
		printf("ERROR: argument number mismatch\n");
	}
	else{

		char* arg = (char*) cmdargs->args;
		dstnode = String_parseUint16(cmdargs->args, NULL);
		printf("Destination Node: %d\n",dstnode);
		arg = ( char* )Cmd_GetNextArg( arg );
		counter = String_parseUint16(arg, NULL);
		printf("Num packets: %d\n",counter);
		arg = ( char* )Cmd_GetNextArg( arg );
		interval = String_parseUint16(arg, NULL);
		printf("Interpacket time: %d\n",interval);

		//open connection to destination node
		OpenTCP(dstnode); 
		
		printf("Connection establishment sent to node %d\n",dstnode);

		//call TCP traffic generator function after 3 secs
		Timers_add(3*1024, TCPtraffic_generator, dstnode);
	}
}



// Software header.
APP(
    APPH_STATIC_APP,	// Application Type
    __APP_NAME,			// Application Name (8 chars)
    1, 0				// Aplication Version (2 bytes)
);
