/*
Copyright 2008, Freie Universitaet Berlin. All rights reserved.

These sources implementing a version of the layer 4 Transmission
Control Protocol (TCP) were developed at the Freie Universitt Berlin, 
for the subject "Telematic Project: Wireless Embedded Systems".

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the authors Joaquin Gonzalez and
Inas Sameer at the subject "Telematic Project: Wireless Embedded Systems"
during the Winter term 2008.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, December 2008
*/


#ifndef TRANSPORT_LAYER_H_
#define TRANSPORT_LAYER_H_

#include <ScatterWeb.System.h>
#include <ScatterWeb.CC1020.h>
#include <string.h>
#include <stdio.h>
#include "ScatterWeb.Process.h"
#include "dsr.h"


/**
 * @brief	TCP Data Segment Type.
 */
#define TCP_DATA_PACKET USERDEFINED_PACKET+5

/**
 * @brief	Retransmission buffer (Send Buffer) size.
 */
#define TAM_BUFF 20

/**
 * @brief	Transmission window size.
 */
#define TAM_TXWNDW 1

/**
 * @brief	TCP Payload Size Definition.
 */
#define TCPPayloadSize 30

/**
 * @brief	Transmission Control Block (TCB) List Size.
 */
#define TCBListSize 10

/**
 * @brief	Maximum number of TCP data retransmissions.
 */
#define DataRetxLimit 8

/**
 * @brief	TCP data retransmission timeout.
 */
#define RTOTCP 12

/**
 * @brief	Maximum number of TCP new data transmission attempts.
 */
#define RetryTCPSend_Limit 3

/**
 * @brief	TCP new data transmission attempt timeout.
 */
#define RetryTCPSend_timeout 3

/**
 * @brief	Maximum number of TCP connection requests (Three-Way Handshake).
 */
#define ConnEstLimit 4

/**
 * @brief	Timeout between consecutive TCP connection requests (Three-Way Handshake).
 */
#define ConnEst_timeout 10

/**
 * @brief	Maximum number of TCP connection termination requests (FIN segment).
 */
#define ConnClosLimit 4

/**
 * @brief	Timeout between consecutive TCP connection termination requests (FIN segment).
 */
#define ConnClos_timeout 2

/**
 * @brief	Timeout before freeing connection resources at initiator node.
 */
#define CloseTCPTimeout 30



//TCP connection states

/**
 @brief	TCP state ST_LISTEN: the server is waiting for an incoming call.
 */
#define ST_LISTEN 0

/**
 @brief	TCP state SYN_SENT: the application has started to open a connection.
 */
#define SYN_SENT 1 

/**
 @brief	TCP state SYN_RCVD: a connection request has arrived, wait for ACK.
 */
#define SYN_RCVD 2

/**
 @brief	TCP state ESTAB: opened connection, normal data transfer.
 */
#define ESTAB 3

/**
 @brief	TCP state FINWAIT1: the application has said the connection is finished.
 */
#define FINWAIT1 4

/**
 @brief	TCP state FINWAIT2: the other side of connection has agreed to release.
 */
#define FINWAIT2 5

/**
 @brief	TCP state TIME_WAIT: wait for all packets to die off.
 */
#define TIME_WAIT 6

/**
 @brief	TCP state CLOSING: both sides have tried to close connection simultaneously.
 */
#define CLOSING 7

/**
 @brief	TCP state CLOSE_WAIT: the other side has initiated release.
 */
#define CLOSE_WAIT 8

/**
 @brief	TCP state LAST_ACK: wait for incoming connection close connection acknowledgement.
 */
#define LAST_ACK 9

/**
 @brief	TCP state CLOSED: no connection is open or pending.
 */
#define CLOSED 10

//TCP Control Byte (C|E|U|A|P|R|S|F)

#define SYN_CB		2		//0000 0010
#define SYNACK_CB	18		//0001 0010
#define ACK_CB		16		//0001 0000
#define PUSH_CB		8		//0000 1000
#define PUSHACK_CB	24		//0001 1000	
#define FIN_CB		1		//0000 0001
#define FINACK_CB	17		//0001 0001


/**
 @brief	TCP state definition.
 */
typedef uint8_t TCP_state_t;


/**
 * @brief	TCP Payload Definition.
 */
typedef uint8_t tcp_payload[TCPPayloadSize];


/**
 * @brief	TCP Segment Definition.
 */
typedef struct{
	uint8_t SEQ_number; ///< Packet Sequence Number
	uint8_t ACK_number; ///< Packet Acknowledgement Number
	uint8_t control; ///< Control Byte (C|E|U|A|P|R|S|F)
	tcp_payload data; ///< Payload Data
}tcp_packet_t;

/**
 * @brief	Send Buffer entry definition.
 */
typedef struct{
	netaddr_t dst; ///< Destination Node Address
	tcp_payload data; ///< Payload Data
	uint8_t SEQ_number; ///< Packet Sequence Number 
	uint8_t retxcount; ///< Data Retransmission counter
	uint8_t inicount;  ///< Initial Retrials counter
	uint8_t sgmnmbr; ///< Segment number as received from application
}sendbuffer_entry_t;

/**
 * @brief	Send Buffer type definition.
 */
typedef sendbuffer_entry_t SendBuffer_t[TAM_BUFF];

/**
 * @brief	Retransmission Buffer (Send Buffer) defined as global variable.
 */
SendBuffer_t SendBuffer; 

/**
 * @brief	Transmission Control Block (TCB) definition.
 */
typedef struct{
	netaddr_t dst; ///< Destination Node Address
	TCP_state_t TCP_state;  ///< Connection state
	uint8_t nextACK_number;   ///< Next expected ACK number
	uint8_t nextSEQ_number;   ///< Next expected SEQ number
	uint8_t TxWindow;	///< Transmission Window
	uint8_t reconncount;   ///< Connection establishment attempts counter
	uint8_t nextSegment;   ///< Next Segment number to be sent after ACK
	uint8_t sgmtcount;   ///<Segment number counter
}TCB_entry_t;

/**
 * @brief	TCB List type definition.
 */
typedef TCB_entry_t TCBList_t[TCBListSize];

/**
 * @brief	TCB List defined as global variable.
 */
TCBList_t TCBList;



//TCP FUNCTIONS DECLARATION

/**
 @brief		Function to initialize TCP protocol resources.
 */
void InitTCP();

/**
 @brief		Function to initialize TCP Retransmission Buffer (Send Buffer).
 */
void InitSendBuffer();

/**
 @brief		Function to initialize TCP Transmission Control Block List.
 */
void InitTCBList();

/**
 @brief				Function to delete TCP Payload data
 @param[in]			data		TCP Payload data
 */
void Clear_TCP_Payload(tcp_payload data);


/**
 @brief				Function to create a TCB entry for a given destination node.
 @param[in]			data		TCP Payload data
 @return			<code>True</code> if the entry was correctly created in TCB Table;
 *					<code>False</code> otherwise.
 */
bool CreateTCBEntry(netaddr_t dstnode);

/**
 @brief				Function to delete a TCB entry for a given destination node.
 @param[in]			data		TCP Payload data
 @return			<code>True</code> if the entry was correctly deleted from TCB Table;
 *					<code>False</code> otherwise.
 */
void DeleteTCBEntry(netaddr_t dstnode);

/**
 @brief				Function to determine the TCB position to a given target node in TCB List
 @param[in]			node		Node address
 @return			TCB position to given node in TCB List.
 */
int Look_up_TCB_Position(netaddr_t node);

/**
 @brief				Function to determine if there is a TCB to a given target node in TCB List
 @param[in]			dstnode		Destination node address
 @return			<code>True</code> if there is an TCB to target node in TCB List;
 *					<code>False</code> otherwise.
 */
bool isConnectionInTCBList(netaddr_t dstnode);

/**
 @brief				Function to obtain the TCB connection state to a given target node in TCB List
 @param[in]			dstnode		Destination node address
 @return			TCB connection state to given target node.
 */
TCP_state_t getTCBState(netaddr_t dstnode);

/**
 @brief				Function that increments the number of connection request attempts of a given TCB
 @param[in]			dstnode		Destination node address
 */
void incTCBConnAttempt(netaddr_t dstnode);

/**
 @brief				Function that obtains the next segment number to be sent from send buffer for a given TCB
 @param[in]			dstnode		Destination node address
 */
uint8_t getTCBSegmNumbr(netaddr_t dstnode);

/**
 @brief				Function that increments the segment number counter to be sent from send buffer for a given TCB
 @param[in]			dstnode		Destination node address
 */
void incTCBSegmCtr(netaddr_t dstnode);

/**
 @brief				Function that obtains the segment number counter to be sent from send buffer for a given TCB
 @param[in]			dstnode		Destination node address
 */
uint8_t getTCBSegmCtr(netaddr_t dstnode);

/**
 @brief				Function that increments the next segment number to be sent from send buffer for a given TCB
 @param[in]			dstnode		Destination node address
 */
void incTCBSegmNumbr(netaddr_t dstnode);


/**
 @brief				Function that obtains the number of connection request attempts so far a given TCB
 @param[in]			dstnode		Destination node address
 */
uint8_t getTCBConnAttempt(netaddr_t dstnode);


/**
 @brief				Function to set the connection state of a target node TCB
 @param[in]			dstnode		Destination node address
 @param[in]			TCBstate	New TCB connection state
 */
void setTCBState(netaddr_t dstnode, TCP_state_t TCBstate);

/**
 @brief				Function to set the ACK number of a TCB for a given target node 
 @param[in]			dstnode		Destination node address
 @param[in]			ACKnumber	New ACK number value
 */
void setTCBACK(netaddr_t dstnode,uint8_t SEQnumber);

/**
 @brief				Function that obtains the SEQ number of a TCB for a given target node
 @param[in]			dstnode		Destination node address
 @return			TCB SEQ number to given target node.
 */
uint8_t getTCBSEQ(netaddr_t dstnode);

/**
 @brief				Function that obtains the ACK number of a TCB for a given target node
 @param[in]			dstnode		Destination node address
 @return			TCB ACK number to given target node.
 */
uint8_t getTCBACK(netaddr_t dstnode);

/**
 @brief				Function that obtains the current Transmission Window Size of a TCB for a given target node
 @param[in]			dstnode		Destination node address
 @return			TCB Transmission Window Size to given target node.
 */
uint8_t getTCBTxWndw(netaddr_t dstnode);

/**
 @brief				Function to set the Transmission Window Size value of a given TCB entry
 @param[in]			dstnode		Destination node address
 @param[in]			newTxWndw	New Transmission Window Size value
 */
void setTCBTxWndw(netaddr_t dstnode, uint8_t newTxWndw);

/**
 @brief				Function to increase the SEQ number of a TCB entry for a given target node
 @param[in]			dstnode		Destination node address
 */
void updateTCBSEQ(netaddr_t dstnode);

/**
 @brief				Function to show the TCB List content
 */
void ShowTCBList();

/**
 @brief				TCP call function to open a connection to a given target node
 @param[in]			dstnode		Destination node address
 */
void OpenTCP(netaddr_t dstnode);

/**
 @brief				TCP call function to close a connection to a given target node
 @param[in]			dstnode		Destination node address
 */
void CloseTCP(netaddr_t dstnode);


/**
 @brief				Function to restart Data Segment Transmission
 @param[in]			*data	Pointer to Send Buffer entry
 */
void Restart_SendTCP(void *datap);

/**
 @brief				Function to send an Acknowledgement segment to a given target node
 @param[in]			*data	Pointer to Send Buffer entry
 */
bool SendACK(netaddr_t dstnode);

/**
 @brief				Function to send a TCP data segment to a given target node
 @param[in]			dstnode		Destination node address
 @return			<code>True</code> if segment was successfully buffered;
 					<code>False</code> otherwise.
 */
bool SendTCP(void *datap);

/**
 @brief				Function to obtain a Send Buffer entry to a given target node and SEQ number
 @param[in]			dst_node	Destination node address
 @param[in]			SEQnumber	Payload Data SEQ number
 @return			Pointer to inserted send buffer entry
 */
sendbuffer_entry_t* getSendBufferEntry(netaddr_t dstnode, uint8_t SEQnum);

/**
 @brief				Function to obtain the next send buffer entry waiting to be sent
 @param[in]			dst_node	Destination node address
 @param[in]			segmentnum	Segment number
 @return			Pointer to inserted send buffer entry
 */
sendbuffer_entry_t* getNextBuffSegm(netaddr_t dstnode);


/**
 @brief				Function to delete a Send Buffer entry to a given target node
 @param[in]			dst_node	Destination node address
 @param[in]			SEQnumber	Payload Data SEQ number
 */
void deleteSendBufferEntry(netaddr_t dstnode, uint8_t SEQnum);

/**
 @brief				Function to check if there are waiting segments to a given target node in Send Buffer
 @param[in]			dst_node	Destination node address
 @return			<code>True</code> if there are waiting segments in Send Buffer;
 					<code>False</code> otherwise.
 */
bool LookupWaitingSegment(netaddr_t dstnode);


/**
 @brief				Function to show the content of a TCP segment
 @param[in]			tcp_pckt		Pointer to TCP segment
 */
void Show_TCP_Packet(tcp_packet_t* tcp_pckt);

/**
 @brief				Function to show the Retransmission Buffer (Send Buffer) content
 */
void ShowSendBuffer();

/**
 @brief				Function that inspects if a segment is a connection request (SYN active)
 @param[in]			tcp_pckt		Pointer to TCP segment
 @return			<code>True</code> if segment is a connection request;
 					<code>False</code> otherwise.
 */
bool isSYNsegment(tcp_packet_t* tcp_pckt);

/**
 @brief				Function that inspects if a segment is a connection response (SYN-ACK active)
 @param[in]			tcp_pckt		Pointer to TCP segment 
 @return			<code>True</code> if segment is a connection response;
 					<code>False</code> otherwise.
 */
bool isSYNACKsegment(tcp_packet_t* tcp_pckt);

/**
 @brief				Function that inspects if a segment is a termination request (FIN active)
 @param[in]			tcp_pckt		Pointer to TCP segment 
 @return			<code>True</code> if segment is a termination request;
 					<code>False</code> otherwise.
 */
bool isFINsegment(tcp_packet_t* tcp_pckt);

/**
 @brief				Function that inspects if a segment is a termination response (FIN-ACK active)
 @param[in]			tcp_pckt		Pointer to TCP segment 
 @return			<code>True</code> if segment is a termination response;
 					<code>False</code> otherwise.
 */
bool isFINACKsegment(tcp_packet_t* tcp_pckt);

/**
 @brief				Function to check if segment has its ACK control bit active
 @param[in]			tcp_pckt		Pointer to TCP segment 
 @return			<code>True</code> if segment has its ACK control bit active;
 					<code>False</code> otherwise.
 */
bool isACKactive(tcp_packet_t* tcp_pckt);

/**
 @brief				Function to check if segment has its PSH control bit active (contains data)
 @param[in]			tcp_pckt		Pointer to TCP segment 
 @return			<code>True</code> if segment has its PSH control bit active;
 					<code>False</code> otherwise.
 */
bool isPSHactive(tcp_packet_t* tcp_pckt);

/**
 @brief				Function to check if a received segment has a valid SEQ number in connection
 @param[in]			dstnode			Destination node address
 @param[in]			rcvdSEQ			SEQ number of received segment
 @return			<code>True</code> if segment SEQ number is valid;
 					<code>False</code> otherwise.
 */
bool isValidSEQ(netaddr_t dstnode, uint8_t rcvdSEQ);

/**
 @brief				Function to check if a received segment acknowledges the reception of a previously sent segment
 @param[in]			dstnode			Destination node address
 @param[in]			rcvdACK			ACK number of received segment
 @return			<code>True</code> if segment ACK number acknowledges a previosly received segment;
 					<code>False</code> otherwise.
 */
bool isnextACK(netaddr_t dstnode, uint8_t rcvdACK);

/**
 @brief				function to initialize a TCP packet
 @param[in]			tcp_pckt		Pointer to TCP packet
 */
void Init_TCP_Pckt(tcp_packet_t* tcp_pckt);

/**
 @brief				function to clear TCP payload data
 @param[in]			data		TCP Payload data
 */
void Clear_TCP_Payload(tcp_payload data);

/**
 @brief				Transport packet handling function
 @param[in]			tcp_pckt		Pointer to TCP segment 
 @param[in]			sndnode			Sender node address
 */
void Handle_Tcp_Data_Pckt(tcp_packet_t* tcp_pckt, uint8_t sndnode);

/**
 @brief				Function to insert a new entry in the Send Buffer
 @param[in]			dst_node	Destination node address
 @param[in]			data		TCP Payload data
 @return			Pointer to inserted send buffer entry
 */
sendbuffer_entry_t* Insert_Entry_SendBuffer(netaddr_t dstnode, tcp_payload data);

//Three Way Handshake procedure functions

/**
 @brief				Function to start Three Way Handshake procedure to a given target node
 @param[in]			dstnode		Destination node address
 @return			<code>True</code> if segment connection request was successfully buffered;
 					<code>False</code> otherwise.
 */
bool Start_Three_Way_Handshake(netaddr_t node);

/**
 @brief				Function to restart Three Way Handshake to a given target node after connection establishment timeout
 @param[in]			*data		Pointer to Send Buffer entry
 */
void Restart_Three_Way_Handshake (void *datap);

/**
 @brief				Function to send a TCP connection request (SYN) to a given target node
 @param[in]			dstnode		Destination node address
 @return			<code>True</code> if segment was successfully buffered;
 					<code>False</code> otherwise.
 */
bool SendTCPConnResp(netaddr_t dstnode);

/**
 @brief				Function to send a TCP connection confirmation (SYN-ACK) to a given target node
 @param[in]			dstnode		Destination node address
 @return			<code>True</code> if segment was successfully buffered;
 					<code>False</code> otherwise.
 */
bool SendTCPConnConf(netaddr_t dstnode);

/**
 @brief				Function to retransmit a TCP data segment
 @param[in]			*data		Pointer to Send Buffer entry
 */
void RetryTCPSend (void *datap);


//Connection termination functions

/**
 @brief				Function to restart close connection procedure for a given target node
 @param[in]			*data		Targer node
 */
void Restart_Close_Connection (void *datap);

/**
 @brief				Function to init Close Connection procedure to given target node
 @param[in]			node		Target node address
 @return			<code>True</code> if FIN segment was successfully buffered;
 					<code>False</code> otherwise.
 */
bool Init_Close_Connection(netaddr_t node);

/**
 @brief				Function to send a termination response back to the initiator in Close Connection procedure
 @param[in]			dstnode		Target node address
 @return			<code>True</code> if FIN ACK segment was successfully buffered;
 					<code>False</code> otherwise.
 */
bool SendTCPFINACK(netaddr_t dstnode);

/**
 @brief				Function to close connection at the initiator after termination timeout
 @param[in]			*datap		Target node address
 */
void CloseSenderConnection(void *datap);

/**
 @brief				Function to prompt test metric results recorded by initiator node
 */
void ShowTCPTestSender();

/**
 @brief				Function to prompt test metric results recorded by receiver node
 */
void ShowTCPTestReceiver();

/**
 @brief				Function to clear all Send Buffer entries to a given target node
 @param[in]			dstnode		Target node address 
 */
void ClearSendBufferEntries(netaddr_t dstnode);

/**
 @brief				Function to remove all active timers to a given target node
 @param[in]			dstnode		Target node address 
 */
void RemoveActiveTimers(netaddr_t dstnode);



#endif /*TRANSPORT_LAYER_H_*/
