/*
Copyright 2007, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2007

$Id: ScatterWeb.Infomem.c 101 2007-10-19 08:15:24Z baar@PCPOOL.MI.FU-BERLIN.DE $
*/


/**
 * @file		ScatterWeb.Infomem.c
 * @ingroup		infomem
 * @brief		MSP430 Infomemory Storage
 * @version		$Ref$
 *
 * $Id: ScatterWeb.Infomem.c 101 2007-10-19 08:15:24Z baar@PCPOOL.MI.FU-BERLIN.DE $
 */
#include <string.h>
#include <ScatterWeb.System.h>

void Infomem_read(void* buffer, uint16_t offset, uint8_t size) {
	uint8_t* address = (uint8_t*)INFOMEM_START + offset;
	memcpy(buffer, address, size);
}

bool Infomem_write(uint16_t offset, uint8_t count, ...) {
	char backup[INFOMEM_BLOCK_SIZE];
	
	register uint8_t* buffer;
	register uint16_t i;
	uint8_t*	 flash;

	// test for address outside infomem
	if( (offset & 0xFF00) > 0 )
		return false;

	// calculate start address of block and in-block offset
	i = offset;
	offset &= INFOMEM_BLOCK_MASK;
	i -= offset;
	flash = (uint8_t*)INFOMEM_START + i;

	//LOG_LOW("block %u blockoffset %u flashing from %p \r\n", i, offset, flash);

	// disable interrupts while writing
	dint();

	// backup into RAM
	memcpy(backup, flash, INFOMEM_BLOCK_SIZE);

	// merge backup with new data
	va_list argp;
	va_start(argp, count);

	buffer = (uint8_t*)backup + offset;
	for( i = 0; i < count; i++ ) {
		register uint16_t	size;
		register uint8_t* data;

		data = va_arg(argp, uint8_t*);
		size = va_arg(argp, uint16_t);
		memcpy(buffer, data, size);
		buffer += size;
	}
	va_end(argp);

	// init flash access
    FCTL2 = FWKEY + FSSEL1 + FN2;
    FCTL3 = FWKEY;

	// erase flash
    FCTL1 = FWKEY + ERASE;
	*flash = 0;

	// write flash
    FCTL1 = FWKEY + WRT;
	buffer = (uint8_t*)backup;
	for( i = 0; i < INFOMEM_BLOCK_SIZE; i++ ) {
		*flash = *buffer;
		buffer++; flash++;
	}
    FCTL1 = FWKEY;
    FCTL3 = FWKEY + LOCK;

	eint();
	return true;
}

#if SYSTEM_CMDLEVEL >= VERBOSE
	/**
	 * @brief	Dump Infomemory: <code>infomem [offset] [length]</code>
	 * @ingroup	commands
	 * @param	offset	Offset to start from (optional)
	 * @param	length	Number of bytes to read (optional)
	 */
	COMMAND(infomem, CMDFLAG_SERIAL, cmdargs) {
		uint8_t* c = (uint8_t*)cmdargs->args;
		uint16_t from = String_parseUint16(c, &c); c++;
		uint16_t length = String_parseUint16(c, &c);

		if( from > 256 ) return;
		if( length == 0 || length > 256 ) length = 256;

		from += (uint16_t)INFOMEM_START;
		System_memoryDump((void*)from, (void*)(from + length));
	}
#endif

